import numpy as np
import pandas as pd
from pathlib import Path


def load_disease():
    file_path = Path(__file__).parent / "data" / "disease.csv"
    df = pd.read_csv(file_path, dtype=np.float_)
    return df



from collections import OrderedDict
from itertools import combinations
from typing import List

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import r2_score

from .gd_regressor import GeoDetectorRegressor


class GeoDetector:
    """ GeoDetector proveid an easy to use interface to calculate the q values
    of the covariates.
    
    References
    ----------
    .. [1] Wang JF, Li XH, Christakos G, Liao YL, Zhang T, Gu X & Zheng XY. 2010.
    Geographical detectors-based health risk assessment and its application in 
    the neural tube defects study of the Heshun region, China. International 
    Journal of Geographical Information Science 24(1): 107-127.
    
    .. [2] Wang JF, Zhang TL, Fu BJ. 2016.A measure of spatial stratified 
    heterogeneity. Ecological Indicators 67(2016): 250-256. 
    """

    def __init__(self,
                 data: pd.DataFrame,
                 x_names: List[str] = "1:",
                 y_name: str = "0",
                 *,
                 method=None,
                 random_state=42):
        self.gd_regressor = GeoDetectorRegressor(method=method,
                                                 random_state=random_state)
        self.x_names = x_names
        self.y_name = y_name
        self.data = data

        if x_names == "1:":
            self.x_names = list(data.columns)[1:]

        if y_name == "0":
            self.y_name = list(data.columns)[0]

        self._check_valid_data(data, self.x_names,
                               self.y_name)  #  type: ignore

    def q_values(self):
        """_q_values calculate the q values of the covariates

        Returns
        -------
        dict, the key is the covariate name, and the value is the q value
        """
        # key is the name of x, value is the q value of x
        dic = OrderedDict()
        for x_name in self.x_names:
            self.gd_regressor.fit(self.data[[x_name]], self.data[self.y_name])
            y_true = self.data[self.y_name]
            # the mean of y in each group is the prediction of y
            y_pred = self.gd_regressor.predict(self.data[[x_name]])
            # 1 - SSW/SST = 1 - MSE/VAR = R2
            q = r2_score(y_true, y_pred)
            dic[x_name] = q
        return dic

    def inter_q_values(self):
        """ factor_detect detect the interaction between covariates
        
        Returns
        -------
        dict, the key is the tuple covariate name, and the value is the q value
        
        Examples
        --------
        
        """
        dic = OrderedDict()
        for x1, x2 in combinations(self.x_names, 2):
            self.gd_regressor.fit(self.data[[x1, x2]], self.data[self.y_name])
            y_true = self.data[self.y_name]
            # the mean of y in each group is the prediction of y
            y_pred = self.gd_regressor.predict(self.data[[x1, x2]])
            # 1 - SSW/SST = 1 - MSE/VAR = R2
            q = r2_score(y_true, y_pred)
            dic[(x1, x2)] = q
        return dic

    def interaction_detect(self, interaction_type=False):
        """ factor_detect detect the interaction between covariates
        Parameters
        ----------
        return_union: bool, default False
        whether return the union of the covariates

        Returns
        -------
        a dataframe, the index and columns are the covariate names, and the 
        value is the q value
        """
        df = pd.DataFrame(index=self.x_names,
                          columns=self.x_names,
                          dtype=np.float_)

        q_dic = self.q_values()
        for x_name in self.x_names:
            df.loc[x_name, x_name] = q_dic[x_name]

        inter_q_dic = self.inter_q_values()
        for cols, q in inter_q_dic.items():
            df.loc[cols[0], cols[1]] = q
            df.loc[cols[1], cols[0]] = q

        if not interaction_type:
            return df
        elif interaction_type:
            interaction_type = pd.DataFrame(index=self.x_names,
                                            columns=self.x_names,
                                            dtype=np.int_)
            for cols, q in inter_q_dic.items():
                new_q = df.loc[cols[0], cols[1]]
                q_0 = q_dic[cols[0]]
                q_1 = q_dic[cols[1]]
                if new_q < min(q_0, q_1):
                    # no linear -
                    interaction_type.loc[cols[0], cols[1]] = 0
                elif min(q_0, q_1) < new_q < max(q_0, q_1):
                    # single_no_linear -
                    interaction_type.loc[cols[0], cols[1]] = 1
                elif max(q_0, q_1) < new_q < q_0 + q_1:
                    # bi+
                    interaction_type.loc[cols[0], cols[1]] = 2
                elif new_q == q_0 + q_1:
                    # alone
                    interaction_type.loc[cols[0], cols[1]] = 3
                elif new_q > q_0 + q_1:
                    # no linear +
                    interaction_type.loc[cols[0], cols[1]] = 4
                else:
                    raise ValueError("new_q is not in the range")
            return df, interaction_type
        else:
            raise ValueError("interaction_type must be bool")

    def plot_interaction(self, show=True):
        interaction_df = self.interaction_detect(interaction_type=False)
        # plot the heatmap of the interaction
        # cbar max value is 1
        ax = sns.heatmap(
            interaction_df,
            annot=True,
            cmap="crest",
            #  vmax=1,
            #  vmin=0,
            linewidths=0.5,
            linecolor="black")

        if show:
            plt.show()
            return ax
        else:
            return ax

    def plot(self, show=True):
        q = self.q_values()
        lst = [(i, j) for i, j in q.items()]
        lst.sort(key=lambda x: x[1], reverse=True)
        key, value = zip(*lst)
        key = np.array(key)
        value = np.array(value)

        ax = sns.barplot(
            x=value,
            y=key,
            orient="h",
            palette="crest",
            width=0.5,
            edgecolor="black",
        )

        # show value in the bar
        for p in ax.patches:
            ax.annotate(
                "%.2f" % p.get_width(),
                (p.get_width() / 2, p.get_y() + p.get_height() / 2.),
                ha="left",
                va="center",
                xytext=(-30, 0),
                textcoords="offset points",
            )

        if show:
            plt.show()
            return ax
        else:
            return ax
        
    def f_value(self):
        dic = OrderedDict()
        for x_name in self.x_names:
            self.gd_regressor.fit(self.data[[x_name]], self.data[self.y_name])
            y_true = self.data[self.y_name]
            # the mean of y in each group is the prediction of y
            y_pred = self.gd_regressor.predict(self.data[[x_name]])
            # 1 - SSW/SST = 1 - MSE/VAR = R2
            q = r2_score(y_true, y_pred)
            dic[x_name] = q
        return dic
        

    def _check_valid_data(self, data: pd.DataFrame, x_names: List[str],
                          y_names: str):
        for x_name in x_names:
            assert x_name in data.columns, f"{x_name} is not in data."

        assert y_names in data.columns, f"{y_names} is not in data."



import numpy as np
import numpy.typing as npt


def groupby(X: npt.NDArray[np.int_], y: npt.NDArray[np.float_]):
    """group y by x

    Parameters
    ----------
    X : an numpy ndarray, shape (n_samples, n_features), dtype int
        the data clustered
    y : an numpy array, shape (n_samples,), dtype float
        the dependent variable
        
    Returns
    -------
    Dict[Union[int, str], numpy array], each key is cluster id, each value is the array
    
    Examples
    --------
    >>> X = np.array([[1, 2], [1, 3], [1, 2],[2, 3],[2, 4],[2, 5]])
    >>> y = np.array([1.7, 2.5, 3.3, 4.1, 5.6, 6.2])
    >>> y_by_x = group_by(X, y)
    >>> y_by_x
    {(1, 2): array([1.7, 3.3]), (1, 3): array([2.5]), (2, 3): array([4.1]), (2, 4): array([5.6]), (2, 5): array([6.2])}
    """
    # sort by each col
    col_num = X.shape[1]

    # index
    idx = np.arange(X.shape[0]).reshape(-1, 1)

    # add index to X
    X_idx = np.hstack((X, idx))

    # sort each col
    for i in range(col_num - 1, -1, -1):
        X_idx = X_idx[X_idx[:, i].argsort(kind="stable")]

    X = X_idx[:, :-1]  # sorted X
    idx = X_idx[:, -1].astype(int)  # sorted index
    y = y[idx]  # sorted y

    # group by each col
    unique_val, unique_idx = np.unique(X, axis=0, return_index=True)
    layer_y = np.split(y, unique_idx)[1:]
    result = dict(zip(map(tuple, unique_val), layer_y))
    return result



import numpy as np
from sklearn.base import BaseEstimator,RegressorMixin
from sklearn.cluster import KMeans
from sklearn.utils.random import check_random_state
from sklearn.utils.validation import check_is_fitted

from .utils import groupby


class GeoDetectorRegressor(RegressorMixin, BaseEstimator):
    """

    The GeoDetector has a sklearn-like interface which can be used to detect
    the spatial stratified heterogeneity of a variable Y,or the determinant
    power of a covariate X of Y.
    The predict method of GeoDetector returns the mean of Y in each subgroup.
    
    Parameters
    ----------
    method: {ClusterMixin, str}, default=None
        method is the method to split x into subgroups.
        method can be a sklearn-like cluster method, or a string pre-defined
        in GeoDetector. such as "kmeans".
        
        if method is None, the dtype of X must be int.
        
        if method is a sklearn-like cluster method, then x will be clustered 
        by the method then the GeoDetector will be applied to the clustered x.
        
        if method is a string, then the pre-defined method will be used to
        cluster x. 
        Pre-defined methods include: kmeans (nature_breaks).
    
    Example:
    >>> from geodetector import GeoDetector
    >>> from geodetector.datasets import load_demo
    >>> gd = GeoDetector()
    >>> gd.fit(load_demo())
    
    References
    ----------
    .. [1] Wang JF, Li XH, Christakos G, Liao YL, Zhang T, Gu X & Zheng XY. 2010.
    Geographical detectors-based health risk assessment and its application in 
    the neural tube defects study of the Heshun region, China. International 
    Journal of Geographical Information Science 24(1): 107-127.
    
    .. [2] Wang JF, Zhang TL, Fu BJ. 2016.A measure of spatial stratified 
    heterogeneity. Ecological Indicators 67(2016): 250-256. 
    """

    def __init__(self, *, method=None, random_state=42):
        # method is the method to split x into subgroups.
        self.method = method
        self.random_state = random_state  # random state

    def fit(self, X, y):
        """fit the model

        Parameters
        ----------
        X : {numpy array, pandas DataFrame}
            the variable to be used to detect the spatial stratified
        y : {numpy array, pandas Series}
            the dependent variable
        """
        random_state = check_random_state(self.random_state)
        X, y = self._validate_data(X, y, reset=True, y_numeric=True)

        if self.method is not None:
            # cluster each col of x when method is not None
            if isinstance(self.method, str):
                self.method = self._get_method(self.method)

            X = self._cluster(X, self.method)

        # assert dtype of x is int
        try:
            X = X.astype(np.int_)
        except ValueError:
            raise ValueError("dtype of X must be int when method is None.")

        assert issubclass(X.dtype.type, np.integer), "dtype of X must be int."

        # group y by x
        # group_val is a dict, key is the layer ,and value is the array of y
        group_val = groupby(X, y)
        judge_dic = {key: val.mean() for key, val in group_val.items()}
        self.judge_dic = judge_dic  # map group to layer_mean

    def predict(self, X):
        """predict the spatial stratified heterogeneity of X

        Parameters
        ----------
        X : {numpy array, pandas DataFrame}
            the variable to be used to detect the spatial stratified

        Returns
        -------
        numpy array
            the spatial stratified heterogeneity of X
        """
        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        if self.method is not None:
            X = self._cluster(X, self.method)
        else:
            X = X.astype(np.int_)

        y = np.array([self.judge_dic[tuple(x)] for x in X])
        return y

    def _cluster(self, X, method):
        """use method to cluster each col of X

        Parameters
        ----------
        X : numpy array 
            the input data to be clustered
        method : sklearn-like cluster method 
            the method to cluster x
            
        Returns
        -------
        numpy array
            the clustered x
        """
        shape = X.shape
        clustered_X = np.zeros(shape)
        for i in range(shape[1]):
            clustered_X[:, [i]] = method.fit_predict(X[:, [i]]).reshape(-1, 1)
        return clustered_X

    def _get_method(self, method):
        method = method.lower()
        if method == "kmeans" or method == "nb":
            return KMeans(random_state=self.random_state, n_clusters=5)
        elif method.startswith("kmeans") or method.startswith("nb"):
            n_clusters = int(method.split("_")[1])
            return KMeans(random_state=self.random_state,
                          n_clusters=n_clusters)
        else:
            raise ValueError(f"method {method} is not supported.")
